# -*- coding: utf-8 -*-

from __future__ import absolute_import
from aiida.orm import CalculationFactory, DataFactory, Code
from aiida.orm.data.base import Float
from aiida.work import WorkChain, workfunction, submit
from aiida.work.workchain import ToContext, Outputs
from copy import deepcopy
from aiida_raspa.workflows import RaspaConvergeWorkChain
from math import sqrt

CifData = DataFactory('cif')
ParameterData = DataFactory('parameter')
SingleFile = DataFactory('singlefile')
SinglefileData = DataFactory('singlefile')
RaspaCalculation = CalculationFactory('raspa')


@workfunction
def update_raspa_parameters(parameters, pressure):
    """Store input parameters of Raspa for given pressure.

    Note: In order to keep the provenance of both Raspa calculations,
    changing the pressure force us to create a new ParameterData node.
    "workfunctions" will take care of linking the user-provided ParameterData
    node to the new one containing the pressure.
    """
    param_dict = parameters.get_dict()
    param_dict['GeneralSettings']['ExternalPressure'] = pressure.value
    return ParameterData(dict=param_dict)


def dict_merge(dct, merge_dct):
    """ Taken from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    import collections
    for k, v in merge_dct.iteritems():
        # This method brakes because component is a list that contains a dict.
        if k == 'Component':
            component_dict = dct['Component'][0]
            for k, v in merge_dct['Component'][0].iteritems():
                component_dict[k] = v
            merge_dct['Component'] = [component_dict]
        elif (k in dct and isinstance(dct[k], dict)
              and isinstance(merge_dct[k], collections.Mapping)):
            dict_merge(dct[k], merge_dct[k])

        else:
            dct[k] = merge_dct[k]


@workfunction
def merge_ParameterData(p1, p2):
    p1_dict = p1.get_dict()
    p2_dict = p2.get_dict()
    dict_merge(p1_dict, p2_dict)
    return ParameterData(dict=p1_dict).store()


class TailCorrectionCheckBlockedPockets(WorkChain):
    """Test the magic of tailcorrections"""

    @classmethod
    def define(cls, spec):
        """Define workflow specification.

        This is the most important method of a Workchain, which defines the
        inputs it takes, the logic of the execution and the outputs
        that are generated in the process.
        """
        super(TailCorrectionCheckBlockedPockets, cls).define(spec)

        # zeopp
        spec.input('zeopp_code', valid_type=Code)
        spec.input("_zeopp_options",
                   valid_type=dict,
                   default=None,
                   required=False)
        spec.input("zeopp_probe_radius", valid_type=Float)
        spec.input("zeopp_atomic_radii",
                   valid_type=SinglefileData,
                   default=None,
                   required=False)

        # First we define the inputs, specifying the type we expect
        spec.input("structure", valid_type=CifData, required=True)
        spec.input("raspa_parameters", valid_type=ParameterData, required=True)
        spec.input("raspa_code", valid_type=Code, required=True)

        # The outline describes the basic logic that defines
        # which steps are executed in what order and based on
        # what conditions. Each `cls.method` is implemented below
        spec.outline(
            cls.init,
            cls.run_zeopp,
            cls.run_raspa_widom,  # for henry coefficient
            cls.run_loading_raspa_low_p,  # for loading at low p
            cls.parse_loading_raspa,
            cls.run_loading_raspa_mid_p,  # for loading at mid p
            cls.parse_loading_raspa,
            cls.run_loading_raspa_high_p,  # for loading at high p
            cls.parse_loading_raspa,
            cls.run_nvt_raspa,  # for rdf and check
            cls.extract_results,  # create dictionary with results
        )

        # Here we define the output the Workchain will generate and
        # return. Dynamic output allows a variety of AiiDA data nodes
        # to be returned
        spec.dynamic_output()

        # Options are used to retrieve RDF if needed

    def init(self):
        """Initialize variables."""

        # dictionaries to save the results for the loading loops
        self.ctx.loading_average = {}
        self.ctx.loading_dev = {}

        self.ctx.gcmc_host_ads_total_energy_average = {}
        self.ctx.gcmc_host_ads_total_energy_dev = {}

        self.ctx.gcmc_host_ads_vdw_energy_average = {}
        self.ctx.gcmc_host_ads_vdw_energy_dev = {}

        self.ctx.gcmc_total_energy_average = {}
        self.ctx.gcmc_total_energy_dev = {}

        self.ctx.gcmc_tail_correction_energy_average = {}
        self.ctx.gcmc_tail_correction_energy_dev = {}

        self.ctx.options = {
            "resources": {
                "num_machines": 1,  # run on 1 node
                "tot_num_mpiprocs": 1,  # use 1 process
                "num_mpiprocs_per_machine": 1,
            },
            "max_wallclock_seconds": 72 * 60 * 60,  # 72 h walltime
            "max_memory_kb": 16000000,  # 16GB memory
            "queue_name": "serial",  # slurm serial partition
            "withmpi": False,  # we run in serial mode
        }

        self.ctx.number_mc_cycles = 15000
        self.ctx.print_every = int(self.ctx.number_mc_cycles) / 10
        self.ctx.number_init_cylces = 15000

    def run_zeopp(self):
        """Main function that performs zeo++ VOLPO and block calculations."""
        params = {
            'ha': True,
            #100 samples / Ang^3: accurate for all the structures
            'block': [self.inputs.zeopp_probe_radius.value, 100],
        }

        inputs = {
            'code': self.inputs.zeopp_code,
            'structure': self.inputs.structure,
            'parameters': NetworkParameters(dict=params).store(),
            '_options': self.inputs._zeopp_options,
            '_label': "ZeoppVolpoBlock",
        }

        # Use default zeopp atomic radii only if a .rad file is not specified
        try:
            inputs['atomic_radii'] = self.inputs.zeopp_atomic_radii
            self.report("Zeopp will use atomic radii from the .rad file")
        except:
            self.report("Zeopp will use default atomic radii")

        # Create the calculation process and launch it
        running = submit(ZeoppCalculation.process(), **inputs)
        self.report(
            "pk: {} | Running zeo++ volpo and block calculations".format(
                running.pid))
        return ToContext(zeopp=Outputs(running))

    def run_raspa_widom(self):
        """Run a Widom calculation in Raspa to get the Henry coefficient"""
        raspa_widom_default_param = {
            "GeneralSettings": {
                'NumberOfInitializationCycles': 0,
                "NumberOfCycles": self.ctx.number_mc_cycles,
                "PrintPropertiesEvery":
                self.ctx.print_every,  # info on henry coeff
                "PrintEvery": self.ctx.print_every,  # info on loading
            },
            "Component": [{
                "WidomProbability": 1.0,
            }],
        }

        bs = self.ctx.zeopp['block']
        with open(bs.get_abs_path() + '/path/' + bs.get_folder_list()[0]) as f:
            self.ctx.number_blocking_spheres = int(f.readline().strip())

        if self.ctx.number_blocking_spheres > 0:
            raspa_widom_default_param['block_component_0'] = bs
            self.report(
                "Blocking spheres ({}) are present and used for Raspa".format(
                    self.ctx.number_blocking_spheres))
        else:
            self.report("No blocking spheres found")

        self.ctx.raspa_widom_parameters = deepcopy(raspa_widom_default_param)
        dict_merge(self.ctx.raspa_widom_parameters,
                   self.inputs.raspa_parameters.get_dict())

        # Create the input dictionary
        inputs = {
            'code':
            self.inputs.raspa_code,
            'structure':
            self.inputs.structure,
            'parameters':
            ParameterData(dict=self.ctx.raspa_widom_parameters),
            '_options':
            self.ctx.options,
            'settings':
            ParameterData(
                dict={
                    'additional_retrieve_list':
                    ['RadialDistributionFunctions/System_0/*'],
                })
        }

        # Create the calculation process and launch it
        running = submit(RaspaConvergeWorkChain, **inputs)
        self.report("pk: {} | Running raspa for the Henry coefficients".format(
            running.pid))

        return ToContext(raspa_widom=Outputs(running))

    def run_loading_raspa_low_p(self):
        self.ctx.current_pressure = 5.8e5
        self.ctx.current_pressure_label = "low"
        return self._run_loading_raspa()

    def run_loading_raspa_mid_p(self):
        self.ctx.current_pressure = 35.0e5
        self.ctx.current_pressure_label = "mid"
        return self._run_loading_raspa()

    def run_loading_raspa_high_p(self):
        self.ctx.current_pressure = 65.0e5
        self.ctx.current_pressure_label = "high"
        return self._run_loading_raspa()

    def _run_loading_raspa(self):
        """Perform RASPA GCMC at given pressure.
        """
        # Create the input dictionary
        raspa_loading_default_params = {
            "GeneralSettings": {
                "NumberOfCycles": self.ctx.number_mc_cycles,
                "NumberOfInitializationCycles": self.ctx.number_init_cylces,
                "PrintEvery": self.ctx.print_every,
            },
            "Component": [{
                "MolFraction": 1.0,
                "TranslationProbability": 1.0,
                "ReinsertionProbability": 1.0,
                "SwapProbability": 1.0,
                "CreateNumberOfMolecules": 0,
            }],
        }

       
        if self.ctx.number_blocking_spheres > 0:
            raspa_loading_default_params['block_component_0'] = bs
            self.report(
                "Blocking spheres ({}) are present and used for Raspa".format(
                    self.ctx.number_blocking_spheres))
        else:
            self.report("No blocking spheres found")

        self.ctx.raspa_loading_parameters = deepcopy(
            raspa_loading_default_params)

        dict_merge(self.ctx.raspa_loading_parameters,
                   self.inputs.raspa_parameters.get_dict())

        inputs = {
            'code':
            self.inputs.raspa_code,
            'structure':
            self.inputs.structure,
            'parameters':
            update_raspa_parameters(
                ParameterData(dict=self.ctx.raspa_loading_parameters),
                Float(self.ctx.current_pressure)),
            '_options':
            self.ctx.options,
            'settings':
            ParameterData(
                dict={
                    'additional_retrieve_list':
                    ['RadialDistributionFunctions/System_0/*'],
                })
        }

        # Create the calculation process and launch it
        running = submit(RaspaConvergeWorkChain, **inputs)
        self.report("pk: {} | Running raspa for the pressure {} [bar]".format(
            running.pid, self.ctx.current_pressure / 1e5))

        return ToContext(raspa=Outputs(running))

    def parse_loading_raspa(self):
        """Extract results from GCMC calculation"""
        loading_average = self.ctx.raspa[
            "component_0"].dict.loading_absolute_average
        loading_dev = self.ctx.raspa["component_0"].dict.loading_absolute_dev
        self.ctx.loading_average[
            self.ctx.current_pressure_label] = loading_average
        self.ctx.loading_dev[self.ctx.current_pressure_label] = loading_dev

        # total host-ads energy
        self.ctx.gcmc_host_ads_total_energy_average[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('host_ads_total_energy_average')
        self.ctx.gcmc_host_ads_total_energy_dev[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('host_ads_total_energy_dev')

        # host-ads vdw
        self.ctx.gcmc_host_ads_vdw_energy_average[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('host_ads_vdw_energy_average')
        self.ctx.gcmc_host_ads_vdw_energy_dev[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('host_ads_vdw_energy_dev')

        # total energy
        self.ctx.gcmc_total_energy_average[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('total_energy_average')
        self.ctx.gcmc_total_energy_dev[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('total_energy_dev')

        # tail correction energy
        self.ctx.gcmc_tail_correction_energy_average[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('tail_correction_energy_average')
        self.ctx.gcmc_tail_correction_energy_dev[
            self.ctx.current_pressure_label] = self.ctx.raspa[
                'output_parameters'].get_attr('tail_correction_energy_dev')

    def run_nvt_raspa(self):

        raspa_nvt_default_params = {
            "GeneralSettings": {
                "NumberOfCycles": self.ctx.number_mc_cycles,
                "NumberOfInitializationCycles": self.ctx.number_init_cylces,
                "PrintEvery": self.ctx.print_every,
            },
            "Component": [{
                "CreateNumberOfMolecules": 1,
                "TranslationProbability": 1.0,
                "ReinsertionProbability": 1.0,
            }]
        }

      
        if self.ctx.number_blocking_spheres > 0:
            raspa_nvt_default_params['block_component_0'] = bs
            self.report(
                "Blocking spheres ({}) are present and used for Raspa".format(
                    self.ctx.number_blocking_spheres))
        else:
            self.report("No blocking spheres found")

        self.ctx.raspa_nvt_parameters = deepcopy(raspa_nvt_default_params)

        dict_merge(self.ctx.raspa_nvt_parameters,
                   self.inputs.raspa_parameters.get_dict())

        inputs = {
            'code':
            self.inputs.raspa_code,
            'structure':
            self.inputs.structure,
            'parameters':
            ParameterData(dict=self.ctx.raspa_nvt_parameters),
            '_options':
            self.ctx.options,
            'settings':
            ParameterData(
                dict={
                    'additional_retrieve_list':
                    ['RadialDistributionFunctions/System_0/*'],
                })
        }

        # Create the calculation process and launch it
        running = submit(RaspaConvergeWorkChain, **inputs)
        self.report("pk: {} | Running raspa NVT".format(running.pid))

        return ToContext(raspa_nvt=Outputs(running))

    def extract_results(self):
        """Extract results of the workflow.

        Attaches the results of the raspa calculation and the initial structure to the outputs.
        """
        cf = self.ctx.raspa[
            "component_0"].dict.conversion_factor_molec_uc_to_cm3stp_cm3

        result = {

            # GCMC properties
            "loading_absolute_average_high_p":
            self.ctx.loading_average["high"] * cf,
            "loading_absolute_dev_high_p":
            self.ctx.loading_dev["high"] * cf,
            "loading_absolute_average_medium_p":
            self.ctx.loading_average["mid"] * cf,
            "loading_absolute_dev_medium_p":
            self.ctx.loading_dev["mid"] * cf,
            "loading_absolute_average_low_p":
            self.ctx.loading_average["low"] * cf,
            "loading_absolute_dev_low_p":
            self.ctx.loading_dev["low"] * cf,
            "gcmc_low_p_host_ads_total_energy_average":
            self.ctx.gcmc_host_ads_total_energy_average['low'],
            "gcmc_low_p_host_ads_total_energy_dev":
            self.ctx.gcmc_host_ads_total_energy_dev['low'],
            "gcmc_low_p_host_ads_vdw_energy_average":
            self.ctx.gcmc_host_ads_vdw_energy_average['low'],
            "gcmc_low_p_host_ads_vdw_energy_dev":
            self.ctx.gcmc_host_ads_vdw_energy_dev['low'],
            "gcmc_low_p_total_energy_average":
            self.ctx.gcmc_total_energy_average['low'],
            "gcmc_low_p_total_energy_dev":
            self.ctx.gcmc_total_energy_dev['low'],
            "gcmc_medium_p_host_ads_total_energy_average":
            self.ctx.gcmc_host_ads_total_energy_average['mid'],
            "gcmc_medium_p_host_ads_total_energy_dev":
            self.ctx.gcmc_host_ads_total_energy_dev['mid'],
            "gcmc_medium_p_host_ads_vdw_energy_average":
            self.ctx.gcmc_host_ads_vdw_energy_average['mid'],
            "gcmc_medium_p_host_ads_vdw_energy_dev":
            self.ctx.gcmc_host_ads_vdw_energy_dev['mid'],
            "gcmc_medium_p_total_energy_average":
            self.ctx.gcmc_total_energy_average['mid'],
            "gcmc_medium_p_total_energy_dev":
            self.ctx.gcmc_total_energy_dev['mid'],
            "gcmc_high_p_host_ads_total_energy_average":
            self.ctx.gcmc_host_ads_total_energy_average['high'],
            "gcmc_high_p_host_ads_total_energy_dev":
            self.ctx.gcmc_host_ads_total_energy_dev['high'],
            "gcmc_high_p_host_ads_vdw_energy_average":
            self.ctx.gcmc_host_ads_vdw_energy_average['high'],
            "gcmc_high_p_host_ads_vdw_energy_dev":
            self.ctx.gcmc_host_ads_vdw_energy_dev['high'],
            "gcmc_high_p_total_energy_average":
            self.ctx.gcmc_total_energy_average['high'],
            "gcmc_high_p_total_energy_dev":
            self.ctx.gcmc_total_energy_dev['high'],
            "gcmc_low_p_tail_correction_energy_average":
            self.ctx.gcmc_tail_correction_energy_average['low'],
            "gcmc_low_p_tail_correction_energy_dev":
            self.ctx.gcmc_tail_correction_energy_dev['low'],
            "gcmc_high_p_tail_correction_energy_average":
            self.ctx.gcmc_tail_correction_energy_average['high'],
            "gcmc_high_p_tail_correction_energy_dev":
            self.ctx.gcmc_tail_correction_energy_dev['high'],
            "gcmc_medium_p_tail_correction_energy_average":
            self.ctx.gcmc_tail_correction_energy_average['mid'],
            "gcmc_medium_p_tail_correction_energy_dev":
            self.ctx.gcmc_tail_correction_energy_dev['mid'],

            # NVT
            "nvt_host_ads_total_energy_average":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'host_ads_total_energy_average'),
            "nvt_host_ads_total_energy_dev":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'host_ads_total_energy_dev'),
            "nvt_host_ads_total_energy_units":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'host_ads_total_energy_unit'),
            "nvt_host_ads_vdw_energy_average":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'host_ads_vdw_energy_average'),
            "nvt_host_ads_vdw_energy_dev":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'host_ads_vdw_energy_dev'),
            "nvt_total_energy_average":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'total_energy_average'),
            "nvt_total_energy_dev":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'total_energy_dev'),
            "nvt_total_energy_units":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'total_energy_units'),
            "nvt_tail_correction_energy_average":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'tail_correction_energy_average'),
            "nvt_tail_correction_energy_dev":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'tail_correction_energy_dev'),
            "nvt_tail_correction_energy_units":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'tail_correction_energy_units'),
            "cell_volume_average":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'cell_volume_average'),
            "cell_volume_dev":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'cell_volume_dev'),
            "cell_volume_units":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'cell_volume_units'),
            "framework_density":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'framework_density'),
            "framework_density_units":
            self.ctx.raspa_nvt['output_parameters'].get_attr(
                'framework_density_units'),

            # conversion factor
            "cf":
            cf,

            # widom calculation
            'henry_coefficient_widom_average':
            self.ctx.raspa_widom["component_0"].dict.henry_coefficient_average,
            'henry_coefficient_widom_dev':
            self.ctx.raspa_widom["component_0"].dict.henry_coefficient_dev,
            'henry_coefficient_widom_units':
            self.ctx.raspa_widom["component_0"].dict.henry_coefficient_units,
            'adsorption_energy_widom_average':
            self.ctx.raspa_widom["component_0"].dict.
            adsorption_energy_widom_average,
            'adsorption_energy_widom_dev':
            self.ctx.raspa_widom["component_0"].dict.
            adsorption_energy_widom_dev,
            'adsorption_energy_widom_units':
            self.ctx.raspa_widom["component_0"].dict.
            adsorption_energy_widom_units,
            'chemical_potential_average':
            self.ctx.raspa_widom["component_0"].dict.
            chemical_potential_average,
            'chemical_potential_dev':
            self.ctx.raspa_widom["component_0"].dict.chemical_potential_dev,
            "widom_host_ads_total_energy_average":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'host_ads_total_energy_average'),
            "widom_host_ads_total_energy_dev":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'host_ads_total_energy_dev'),
            "widom_host_ads_total_energy_units":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'host_ads_total_energy_unit'),
            "widom_host_ads_vdw_energy_average":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'host_ads_vdw_energy_average'),
            "widom_host_ads_vdw_energy_dev":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'host_ads_vdw_energy_dev'),
            "widom_total_energy_average":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'total_energy_average'),
            "widom_total_energy_dev":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'total_energy_dev'),
            "widom_total_energy_units":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'total_energy_units'),
            "widom_tail_correction_energy_average":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'tail_correction_energy_average'),
            "widom_tail_correction_energy_dev":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'tail_correction_energy_dev'),
            "widom_tail_correction_energy_units":
            self.ctx.raspa_widom['output_parameters'].get_attr(
                'tail_correction_energy_units'),
        }

        self.out("result", ParameterData(dict=result).store())

        self.report("Workchain <{}> completed successfully".format(
            self.calc.pk))
        return
